
import ydpg
import machine

import math
import time
import json

class WS2812:
	WIDTH = 32
	HEIGHT = 8
	
	WIDTH2 = WIDTH * 2
	PIXELS = WIDTH2 * HEIGHT
	COLORS = PIXELS * 3

	COLORS_VIEW = COLORS // 2
	
	BITS = COLORS_VIEW * 8 + 1 # + additional delay

	@staticmethod
	def copy_color(frame, i1, i2):
		frame[i1] = frame[i2]
		frame[i1+1] = frame[i2+1]
		frame[i1+2] = frame[i2+2]
		
	def copy_bottom(self, i1, i2):
		frame = self.frame
		frame_bottom = self.frame_bottom
		frame[i1] = frame_bottom[i2]
		frame[i1+1] = frame_bottom[i2+1]
		frame[i1+2] = frame_bottom[i2+2]
	
	def __init__(self, pin="tx"):
		self.dim = 0
	
		# create PWM timer & channel
		self.pwm = machine.PWM(pin, prescaler=7, period=49, duty=0)

		# frame buffer
		self.frame = bytearray(WS2812.COLORS)
		self.frame_bottom = bytearray(WS2812.COLORS)
		self.frame_mv = memoryview(self.frame)[WS2812.COLORS_VIEW:]
		# PWM duty cycle buffer
		self.pwm_buf = bytearray(WS2812.BITS)
		
		self.last = time.ticks_us()
		self.fps = 30
	
	def shift(self):
		frame = self.frame
		for x in range(WS2812.WIDTH2 - 1, 0, -1):
			xo = x * WS2812.HEIGHT * 3
			for y in range(WS2812.HEIGHT):
				i1 = xo + y * 3
				i2 = xo - (y + 1) * 3
				
				WS2812.copy_color(frame, i1, i2)
		
		for y in range(WS2812.HEIGHT * 3):
			frame[y] = 0
	
	def scroll(self):
		H1 = WS2812.HEIGHT - 1
		frame = self.frame
		frame_bottom = self.frame_bottom
		for x in range(WS2812.WIDTH):
			o = WS2812.COLORS_VIEW + x * WS2812.HEIGHT * 3
			
			if (x & 0x1) != 0:
				for y in range(H1):
					yo = o + y * 3
					WS2812.copy_color(frame, yo, yo + 3)
				
				self.copy_bottom(o + H1*3, o)
				
				for y in range(H1):
					yo = o + y * 3
					WS2812.copy_color(frame_bottom, yo, yo + 3)
				
				yo = o + H1 * 3
				frame_bottom[yo] = 0
				frame_bottom[yo+1] = 0
				frame_bottom[yo+2] = 0
			else:
				for y in range(H1, 0, -1):
					yo = o + y * 3
					WS2812.copy_color(frame, yo, yo - 3)
					
				self.copy_bottom(o, o + H1*3)
				
				for y in range(H1, 0, -1):
					yo = o + y * 3
					WS2812.copy_color(frame_bottom, yo, yo - 3)
					
				frame_bottom[o] = 0
				frame_bottom[o+1] = 0
				frame_bottom[o+2] = 0
	
	def pixel(self, x, y, color):
		if y < 0:
			y += WS2812.HEIGHT
			frame = self.frame_bottom
		else:
			frame = self.frame
	
		if (x & 0x1) != 0: # 'snake' pattern
			y = (WS2812.HEIGHT - 1) - y
		i = WS2812.COLORS_VIEW + (x * WS2812.HEIGHT + y) * 3
		
		frame[i] = (color & 0xFF) >> self.dim
		frame[i+1] = ((color >> 8) & 0xFF) >> self.dim
		frame[i+2] = ((color >> 16) & 0xFF) >> self.dim

	def fill(self, color):
		G = (color & 0xFF) >> self.dim
		R = ((color >> 8) & 0xFF) >> self.dim
		B = ((color >> 16) & 0xFF) >> self.dim
		frame = self.frame
		for I in range(WS2812.COLORS_VIEW, WS2812.COLORS, 3):
			frame[I] = G
			frame[I+1] = R
			frame[I+2] = B
			
	def clear(self):
		frame = self.frame
		for I in range(WS2812.COLORS_VIEW, WS2812.COLORS):
			frame[I] = 0

	def update(self, timeout=200):
		# Wait for PWM transfer to complete.
		self.pwm.waittx(timeout) 
		# Convert color buffer to PWM duty cycle buffer.
		ydpg.bit2pwm(self.pwm_buf, 0, self.frame_mv, 27, 8)
		# Start PWM transfer.
		# Warning: don't modify 'pwm_buf' until transfer is complete
		self.pwm.send(self.pwm_buf)
		
		now = time.ticks_us()
		self.fps = (self.fps + (1000000 / time.ticks_diff(now, self.last))) * 0.5
		self.last = now

class WSFont:
	@staticmethod
	def readimg(path="font.bmp", width=133, height=16):
		print("Loading font image:", path, width, height)
		
		ret = bytearray(width * math.ceil(height / 8))
		
		with open(path, "rb") as f:
			rgb = bytearray(3)
			# skip header
			f.seek(54)
			
			# BMP pads every row to align on 4-byte boundary
			pad = (width * 3) % 4
			pad = (4 - pad) % 4

			for y in range(height):
				for x in range(width):
					n = f.readinto(rgb)
					if n != 3:
						print("Something went wrong loading image")
						break
					
					avg = (rgb[0] + rgb[1] + rgb[2]) // 3
					avg = 1 if (avg > 127) else 0
					y2 = height - 1 - y
					i = x + (y2 // 8) * width
					ret[i] |= avg << (y2 % 8)
				_ = f.read(pad)
		return ret
	
	def getba(self, c):
		c -= 0x20
		if c >= 65 and c <= 90 and self.chars[c] == None:
			return self.getba(c) # to upper case if missing
		if c < 0 or c > 0x5F or self.chars[c] == None:
			return self.getba(0x7F) # unknown or unprintable character
		return self.chars[c]
	
	def calcw(self, s):
		w = 0
		if len(s) > 0:
			for c in s:
				w += self.getba(ord(c))[0]
			w += (len(s) - 1) * self.gap
		return w
		
	def makechar(self, img, c, xo, yo, w):
		c -= 0x20
		
		ba = bytearray(1 + w * math.ceil(self.height / 8))
		ba[0] = w
		o = xo + (yo // self.height) * self.iwidth
		for x in range(w):
			ba[1 + x] = img[o + x]
		
		self.chars[c] = ba	
	
	def __init__(self, ws, path="font.json"):
		with open(path, "rt") as f:
			obj = json.load(f)
		
		self.ws = ws
		self.height = obj["height"]
		self.gap = obj["gap"]
		self.chars = [None] * 0x60
		
		img_info = obj["img"]
		self.iwidth = img_info["width"]
		self.iheight = img_info["height"]
		
		img = WSFont.readimg(img_info["path"], self.iwidth, self.iheight)
		for c in obj["chars"]:
			c[0] = ord(c[0])
			self.makechar(img, *c)
		
		self.makechar(img, 0x7F, *obj["missing"])
		#self.printdebug()
				
	def putc(self, x0, y0, c, fg, bg=None):
		c = self.getba(c)
		for y in range(self.height):
			y2 = self.height - 1 - y
			yc = 1 + math.floor(y2 / 8) * c[0]
			yp = y2 % 8
			
			for x in range(c[0]):
				b = (c[x + yc] >> yp) & 0x1
				if b != 0:
					self.ws.pixel(x0 - x, y0 + y, fg)
				elif bg != None:
					self.ws.pixel(x0 - x, y0 + y, bg)
			
			if bg != None:
				for x in range(x+1, x+1+self.gap):
					self.ws.pixel(x0 - x, y0 + y, bg)
		#return x + 1
		return x0 - c[0] - self.gap
		
	def puts(self, x, y, s, fg, bg=None):
		for c in s:
			x = self.putc(x, y, ord(c), fg, bg)
		return x
