
import time
from ws2812 import *

COLOR_YD = 0xB03699
COLOR_WHITE = 0xFFFFFF

TYPE_FRAME = 0
TYPE_TICKER = 1
TYPE_CUSTOM = 2

ANIM_NONE = 0
ANIM_VERTICAL = 1
ANIM_HORIZONTAL = 2

# "TICKER" variables
ticker_text = None
ticker_ch = 0
ticker_hidden = 0
ticker_color = 0
		
def ticker_put(font):
	global ticker_text, ticker_ch, ticker_hidden, ticker_color
	c = ord(ticker_text[ticker_ch])
	ticker_ch = (ticker_ch + 1) % len(ticker_text)
	w = font.getba(c)[0]
	font.putc(-1, 0, c, ticker_color)
	ticker_hidden = w + font.gap

def ticker_start(ws, font, cfg):
	global COLOR_YD, ticker_text, ticker_ch, ticker_hidden, ticker_color
	ticker_ch = 0
	
	ticker_text = cfg.get("text", "Yacht Devices")
	ticker_text += " " * cfg.get("spaces", 8);
	ticker_color = cfg.get("color", COLOR_YD)
	ticker_put(font)
	return True
	
def ticker_update(ws, font):
	global ticker_ch, ticker_hidden
	ws.shift()
	ticker_hidden -= 1
	if ticker_hidden <= 0:
		if ticker_ch == 0:
			return None
			
		ticker_put(font)
	return True
	
# "FRAME" variables

frame_vertical = True
frame_hold = 0
frame_hidden = 0

def frame_draw(ws, font, cfg, x, y):
	global COLOR_WHITE
	text_left = cfg.get("text", None)
	text_center = cfg.get("ctext", None)
	text_right = cfg.get("rtext", None)
	draw_count = 0
	
	if text_left != None:
		color = cfg.get("color", COLOR_WHITE)
		font.puts(x + WS2812.WIDTH - 1, y, text_left, color)
		#draw_count += 1
	if text_center != None:
		color = cfg.get("ccolor", COLOR_WHITE)
		o = font.calcw(text_center)
		font.puts(x + (WS2812.WIDTH + o) // 2, y, text_center, color)
		draw_count += 1
	if text_right != None:
		color = cfg.get("rcolor", COLOR_WHITE)
		o = font.calcw(text_right)
		font.puts(x + o - 1, y, text_right, color)
		draw_count += 1
	
	draw = cfg["draw"]
	if draw != None and draw(ws, font, cfg, x, y):
		draw_count += 1
	return draw_count > 0

def frame_start(ws, font, cfg):
	global frame_vertical, frame_hold, frame_hidden
	
	if cfg.hold:
		anim = ANIM_NONE
		frame_hold = 100 / cfg.interval
	else:
		anim = cfg.get("anim", ANIM_HORIZONTAL)
		frame_hold = round(cfg.get("hold", 1) * (1000 / cfg.interval))
		
	if anim == ANIM_VERTICAL:
		frame_vertical = True
		frame_hidden = WS2812.HEIGHT
		ret = frame_draw(ws, font, cfg, 0, -WS2812.HEIGHT)
	elif anim == ANIM_HORIZONTAL:
		frame_vertical = False
		frame_hidden = WS2812.WIDTH
		ret = frame_draw(ws, font, cfg, -WS2812.WIDTH, 0)
	else:
		if cfg.hold: ws.fill(0x000000)
		frame_vertical = None
		frame_hidden = 0
		ret = frame_draw(ws, font, cfg, 0, 0)
	return ret
	
def frame_update(ws, font):
	global frame_hidden, frame_hold, frame_vertical
	if frame_hidden > 0:
		if frame_vertical:
			ws.scroll()
		else:
			ws.shift()
	elif frame_hidden <= 0:
		frame_hold -= 1
		return False if frame_hold >= 0 else None
	frame_hidden -= 1
	return True

class WSConfig:
	def __init__(self, pin, ws, font, config):
		self.pin = pin
		self.ws = ws
		self.font = font
		self.config = config
		
		self.btn_value = 1
		self.btn_last = None
		self.hold = False

		self.active = 0
		self.next_index = None
		self.conf = None
		self.update = None
		self.repeat = 0
		self.elapsed = 0
		self.interval = 1000 / 20
		self.last = time.ticks_ms()

	def __getitem__(self, key):
		return self.conf.get(key, None)

	def get(self, key, default):
		ret = self.conf.get(key, default)
		return ret() if callable(ret) else ret
	
	def select(self, index):
		print("FPS: " + str(self.ws.fps))
		if not self.hold and self.repeat <= 0:
			if self.conf != None and self.conf.get("clear_after", False):
				self.ws.fill(0x000000)
		
			self.active = index % len(self.config)
			self.conf = self.config[self.active]
			
			type_id = self.conf.get("type", TYPE_FRAME)
			if type_id == TYPE_FRAME:
				self.start_func = frame_start
				self.update_func = frame_update
			elif type_id == TYPE_TICKER:
				self.start_func = ticker_start
				self.update_func = ticker_update
			elif type_id == TYPE_CUSTOM:
				self.start_func = self.conf["start"]
				self.update_func = self.conf["update"]
			else:
				self.start_func = None
				self.update_func = None
				
			if self.start_func == None:
				return None
			
			self.repeat = self.conf.get("repeat", 0)
			
			self.next_index = self.active + self.conf.get("next", 1)
			self.interval = 1000 / self.conf.get("fps", 21)
			self.ws.dim = self.conf.get("dim", 2)
		else:
			self.repeat -= 1
			
		if self.conf.get("clear", False): self.ws.clear()
		return self.start_func(self.ws, self.font, self)
	
	def next_frame(self, ret=None):
		while ret == None: ret = self.select(self.next_index)
		if ret: self.ws.update()
		return ret
	
	def start(self):
		try:
			self.last = time.ticks_ms()
			self.next_index = 0
			
			self.next_frame()
			
			while True:
				if self.elapsed >= self.interval:
					self.elapsed = (self.elapsed - self.interval) % self.interval
					ret = self.update_func and self.update_func(self.ws, self.font)
					self.next_frame(ret)

				now = time.ticks_ms()
				diff = time.ticks_diff(now, self.last)
				self.last = now
				self.elapsed += diff
				
				btn = self.pin()
				if btn != self.btn_value:
					self.btn_value = btn
					if self.btn_last or time.ticks_diff(now, self.btn_last) > 100:
						self.btn_last = now
						if btn == 0: # pressed
							print("Pressed!")
							self.hold = not self.hold
		except KeyboardInterrupt:
			#pass
			self.ws.clear()
			self.ws.update()